@@ -33,8 +33,8 @@ const struct cmd_entry cmd_select_pane_entry = {
 	.name = "select-pane",
 	.alias = "selectp",
 
-	.args = { "DdegLlMmP:RT:t:U", 0, 0 },
-	.usage = "[-DdegLlMmRU] [-P style] [-T title] " CMD_TARGET_PANE_USAGE,
+	.args = { "DdegLlMmP:RT:t:U", 0, 0 }, /* -P and -g deprecated */
+	.usage = "[-DdeLlMmRU] [-T title] " CMD_TARGET_PANE_USAGE,
 
 	.target = { 't', CMD_FIND_PANE, 0 },
 
@@ -66,7 +66,7 @@ cmd_select_pane_redraw(struct window *w)
 	 */
 
 	TAILQ_FOREACH(c, &clients, entry) {
-		if (c->session == NULL)
+		if (c->session == NULL || (c->flags & CLIENT_CONTROL))
 			continue;
 		if (c->session->curw->window == w && tty_window_bigger(&c->tty))
 			server_redraw_client(c);
@@ -90,9 +90,10 @@ cmd_select_pane_exec(struct cmd *self, struct cmdq_item *item)
 	struct window		*w = wl->window;
 	struct session		*s = item->target.s;
 	struct window_pane	*wp = item->target.wp, *lastwp, *markedwp;
-	struct style		*sy = &wp->style;
 	char			*pane_title;
 	const char		*style;
+	struct style		*sy;
+	struct options_entry	*o;
 
 	if (self->entry == &cmd_last_pane_entry || args_has(args, 'l')) {
 		lastwp = w->last;
@@ -112,7 +113,7 @@ cmd_select_pane_exec(struct cmd *self, struct cmdq_item *item)
 		else {
 			server_unzoom_window(w);
 			window_redraw_active_switch(w, lastwp);
-			if (window_set_active_pane(w, lastwp)) {
+			if (window_set_active_pane(w, lastwp, 1)) {
 				cmd_find_from_winlink(current, wl, 0);
 				cmd_select_pane_redraw(w);
 			}
@@ -144,15 +145,20 @@ cmd_select_pane_exec(struct cmd *self, struct cmdq_item *item)
 
 	if (args_has(self->args, 'P') || args_has(self->args, 'g')) {
 		if ((style = args_get(args, 'P')) != NULL) {
-			style_set(sy, &grid_default_cell);
-			if (style_parse(sy, &grid_default_cell, style) == -1) {
+			o = options_set_style(wp->options, "window-style", 0,
+			    style);
+			if (o == NULL) {
 				cmdq_error(item, "bad style: %s", style);
 				return (CMD_RETURN_ERROR);
 			}
-			wp->flags |= PANE_REDRAW;
+			options_set_style(wp->options, "window-active-style", 0,
+			    style);
+			wp->flags |= (PANE_REDRAW|PANE_STYLECHANGED);
 		}
-		if (args_has(self->args, 'g'))
+		if (args_has(self->args, 'g')) {
+			sy = options_get_style(wp->options, "window-style");
 			cmdq_print(item, "%s", style_tostring(sy));
+		}
 		return (CMD_RETURN_NORMAL);
 	}
 
@@ -194,9 +200,9 @@ cmd_select_pane_exec(struct cmd *self, struct cmdq_item *item)
 		return (CMD_RETURN_NORMAL);
 	server_unzoom_window(wp->window);
 	window_redraw_active_switch(w, wp);
-	if (window_set_active_pane(w, wp)) {
+	if (window_set_active_pane(w, wp, 1)) {
 		cmd_find_from_winlink_pane(current, wl, wp, 0);
-		hooks_insert(s->hooks, item, current, "after-select-pane");
+		cmdq_insert_hook(s, item, current, "after-select-pane");
 		cmd_select_pane_redraw(w);
 	}
 
