@@ -39,9 +39,10 @@ const struct cmd_entry cmd_split_window_entry = {
 	.name = "split-window",
 	.alias = "splitw",
 
-	.args = { "bc:dfF:l:hp:Pt:v", 0, -1 },
-	.usage = "[-bdfhvP] [-c start-directory] [-F format] "
-		 "[-p percentage|-l size] " CMD_TARGET_PANE_USAGE " [command]",
+	.args = { "bc:de:fF:hIl:p:Pt:v", 0, -1 },
+	.usage = "[-bdefhIPv] [-c start-directory] [-e environment] "
+		 "[-F format] [-p percentage|-l size] " CMD_TARGET_PANE_USAGE
+		 " [command]",
 
 	.target = { 't', CMD_FIND_PANE, 0 },
 
@@ -52,111 +53,106 @@ const struct cmd_entry cmd_split_window_entry = {
 static enum cmd_retval
 cmd_split_window_exec(struct cmd *self, struct cmdq_item *item)
 {
-	struct cmd_find_state	*current = &item->shared->current;
 	struct args		*args = self->args;
+	struct cmd_find_state	*current = &item->shared->current;
+	struct spawn_context	 sc;
 	struct client		*c = cmd_find_client(item, NULL, 1);
 	struct session		*s = item->target.s;
 	struct winlink		*wl = item->target.wl;
-	struct window		*w = wl->window;
-	struct window_pane	*wp = item->target.wp, *new_wp = NULL;
-	struct environ		*env;
-	const char		*cmd, *path, *shell, *template, *tmp;
-	char		       **argv, *cause, *new_cause, *cp, *cwd;
-	u_int			 hlimit;
-	int			 argc, size, percentage, before;
+	struct window_pane	*wp = item->target.wp, *new_wp;
 	enum layout_type	 type;
 	struct layout_cell	*lc;
-	struct environ_entry	*envent;
-	struct cmd_find_state    fs;
+	struct cmd_find_state	 fs;
+	int			 size, percentage, flags, input;
+	const char		*template, *add;
+	char			*cause, *cp;
+	struct args_value	*value;
 
-	server_unzoom_window(w);
-
-	if (args->argc == 0) {
-		cmd = options_get_string(s->options, "default-command");
-		if (cmd != NULL && *cmd != '\0') {
-			argc = 1;
-			argv = (char **)&cmd;
-		} else {
-			argc = 0;
-			argv = NULL;
-		}
-	} else {
-		argc = args->argc;
-		argv = args->argv;
-	}
-
-	if ((tmp = args_get(args, 'c')) != NULL)
-		cwd = format_single(item, tmp, c, s, NULL, NULL);
-	else
-		cwd = xstrdup(server_client_get_cwd(item->client, s));
-
-	type = LAYOUT_TOPBOTTOM;
 	if (args_has(args, 'h'))
 		type = LAYOUT_LEFTRIGHT;
-	before = args_has(args, 'b');
-
-	size = -1;
+	else
+		type = LAYOUT_TOPBOTTOM;
 	if (args_has(args, 'l')) {
 		size = args_strtonum(args, 'l', 0, INT_MAX, &cause);
 		if (cause != NULL) {
-			xasprintf(&new_cause, "size %s", cause);
+			cmdq_error(item, "create pane failed: -l %s", cause);
 			free(cause);
-			cause = new_cause;
-			goto error;
+			return (CMD_RETURN_ERROR);
 		}
 	} else if (args_has(args, 'p')) {
 		percentage = args_strtonum(args, 'p', 0, INT_MAX, &cause);
 		if (cause != NULL) {
-			xasprintf(&new_cause, "percentage %s", cause);
+			cmdq_error(item, "create pane failed: -p %s", cause);
 			free(cause);
-			cause = new_cause;
-			goto error;
+			return (CMD_RETURN_ERROR);
 		}
 		if (type == LAYOUT_TOPBOTTOM)
 			size = (wp->sy * percentage) / 100;
 		else
 			size = (wp->sx * percentage) / 100;
-	}
-	hlimit = options_get_number(s->options, "history-limit");
-
-	shell = options_get_string(s->options, "default-shell");
-	if (*shell == '\0' || areshell(shell))
-		shell = _PATH_BSHELL;
-
-	lc = layout_split_pane(wp, type, size, before, args_has(args, 'f'));
-	if (lc == NULL) {
-		cause = xstrdup("pane too small");
-		goto error;
-	}
-	new_wp = window_add_pane(w, wp, before, args_has(args, 'f'), hlimit);
-	layout_make_leaf(lc, new_wp);
-
-	path = NULL;
-	if (item->client != NULL && item->client->session == NULL)
-		envent = environ_find(item->client->environ, "PATH");
-	else
-		envent = environ_find(s->environ, "PATH");
-	if (envent != NULL)
-		path = envent->value;
-
-	env = environ_for_session(s, 0);
-	if (window_pane_spawn(new_wp, argc, argv, path, shell, cwd, env,
-	    s->tio, &cause) != 0) {
-		environ_free(env);
-		goto error;
-	}
-	environ_free(env);
-
-	layout_fix_panes(w);
-	server_redraw_window(w);
-
-	if (!args_has(args, 'd')) {
-		window_set_active_pane(w, new_wp);
-		session_select(s, wl->idx);
-		cmd_find_from_session(current, s, 0);
-		server_redraw_session(s);
 	} else
-		server_status_session(s);
+		size = -1;
+
+	server_unzoom_window(wp->window);
+	input = (args_has(args, 'I') && args->argc == 0);
+
+	flags = 0;
+	if (args_has(args, 'b'))
+		flags |= SPAWN_BEFORE;
+	if (args_has(args, 'f'))
+		flags |= SPAWN_FULLSIZE;
+	if (input || (args->argc == 1 && *args->argv[0] == '\0'))
+		flags |= SPAWN_EMPTY;
+
+	lc = layout_split_pane(wp, type, size, flags);
+	if (lc == NULL) {
+		cmdq_error(item, "no space for new pane");
+		return (CMD_RETURN_ERROR);
+	}
+
+	memset(&sc, 0, sizeof sc);
+	sc.item = item;
+	sc.s = s;
+	sc.wl = wl;
+
+	sc.wp0 = wp;
+	sc.lc = lc;
+
+	sc.name = NULL;
+	sc.argc = args->argc;
+	sc.argv = args->argv;
+	sc.environ = environ_create();
+
+	add = args_first_value(args, 'e', &value);
+	while (add != NULL) {
+		environ_put(sc.environ, add);
+		add = args_next_value(&value);
+	}
+
+	sc.idx = -1;
+	sc.cwd = args_get(args, 'c');
+
+	sc.flags = flags;
+	if (args_has(args, 'd'))
+		sc.flags |= SPAWN_DETACHED;
+
+	if ((new_wp = spawn_pane(&sc, &cause)) == NULL) {
+		layout_close_pane(new_wp);
+		cmdq_error(item, "create pane failed: %s", cause);
+		free(cause);
+		return (CMD_RETURN_ERROR);
+	}
+	if (input && window_pane_start_input(new_wp, item, &cause) != 0) {
+		layout_close_pane(new_wp);
+		window_remove_pane(wp->window, new_wp);
+		cmdq_error(item, "%s", cause);
+		free(cause);
+		return (CMD_RETURN_ERROR);
+	}
+	if (!args_has(args, 'd'))
+		cmd_find_from_winlink_pane(current, wl, new_wp, 0);
+	server_redraw_window(wp->window);
+	server_status_session(s);
 
 	if (args_has(args, 'P')) {
 		if ((template = args_get(args, 'F')) == NULL)
@@ -165,22 +161,12 @@ cmd_split_window_exec(struct cmd *self, struct cmdq_item *item)
 		cmdq_print(item, "%s", cp);
 		free(cp);
 	}
-	notify_window("window-layout-changed", w);
 
 	cmd_find_from_winlink_pane(&fs, wl, new_wp, 0);
-	hooks_insert(s->hooks, item, &fs, "after-split-window");
+	cmdq_insert_hook(s, item, &fs, "after-split-window");
 
-	free(cwd);
+	environ_free(sc.environ);
+	if (input)
+		return (CMD_RETURN_WAIT);
 	return (CMD_RETURN_NORMAL);
-
-error:
-	if (new_wp != NULL) {
-		layout_close_pane(new_wp);
-		window_remove_pane(w, new_wp);
-	}
-	cmdq_error(item, "create pane failed: %s", cause);
-	free(cause);
-
-	free(cwd);
-	return (CMD_RETURN_ERROR);
 }
