@@ -1,7 +1,7 @@
 /* $OpenBSD$ */
 
 /*
- * Copyright (c) 2012 Nicholas Marriott <nicm@users.sourceforge.net>
+ * Copyright (c) 2012 Nicholas Marriott <nicholas.marriott@gmail.com>
  * Copyright (c) 2012 George Nachman <tmux@georgester.com>
  *
  * Permission to use, copy, modify, and distribute this software for any
@@ -37,7 +37,7 @@ control_write(struct client *c, const char *fmt, ...)
 	va_end(ap);
 
 	evbuffer_add(c->stdout_data, "\n", 1);
-	server_push_stdout(c);
+	server_client_push_stdout(c);
 }
 
 /* Write a buffer, adding a terminal newline. Empties buffer. */
@@ -46,16 +46,31 @@ control_write_buffer(struct client *c, struct evbuffer *buffer)
 {
 	evbuffer_add_buffer(c->stdout_data, buffer);
 	evbuffer_add(c->stdout_data, "\n", 1);
-	server_push_stdout(c);
+	server_client_push_stdout(c);
+}
+
+/* Control error callback. */
+static enum cmd_retval
+control_error(struct cmdq_item *item, void *data)
+{
+	struct client	*c = item->client;
+	char		*error = data;
+
+	cmdq_guard(item, "begin", 1);
+	control_write(c, "parse error: %s", error);
+	cmdq_guard(item, "error", 1);
+
+	free(error);
+	return (CMD_RETURN_NORMAL);
 }
 
 /* Control input callback. Read lines and fire commands. */
 void
-control_callback(struct client *c, int closed, unused void *data)
+control_callback(struct client *c, int closed, __unused void *data)
 {
-	char		*line, *cause;
-	struct cmd_list	*cmdlist;
-	struct cmd	*cmd;
+	char			*line;
+	struct cmdq_item	*item;
+	struct cmd_parse_result	*pr;
 
 	if (closed)
 		c->flags |= CLIENT_EXIT;
@@ -65,24 +80,25 @@ control_callback(struct client *c, int closed, unused void *data)
 		if (line == NULL)
 			break;
 		if (*line == '\0') { /* empty line exit */
+			free(line);
 			c->flags |= CLIENT_EXIT;
 			break;
 		}
 
-		if (cmd_string_parse(line, &cmdlist, NULL, 0, &cause) != 0) {
-			c->cmdq->time = time(NULL);
-			c->cmdq->number++;
-
-			cmdq_guard(c->cmdq, "begin", 1);
-			control_write(c, "parse error: %s", cause);
-			cmdq_guard(c->cmdq, "error", 1);
-
-			free(cause);
-		} else {
-			TAILQ_FOREACH(cmd, &cmdlist->list, qentry)
-				cmd->flags |= CMD_CONTROL;
-			cmdq_run(c->cmdq, cmdlist, NULL);
-			cmd_list_free(cmdlist);
+		pr = cmd_parse_from_string(line, NULL);
+		switch (pr->status) {
+		case CMD_PARSE_EMPTY:
+			break;
+		case CMD_PARSE_ERROR:
+			item = cmdq_get_callback(control_error, pr->error);
+			cmdq_append(c, item);
+			break;
+		case CMD_PARSE_SUCCESS:
+			item = cmdq_get_command(pr->cmdlist, NULL, NULL, 0);
+			item->shared->flags |= CMDQ_SHARED_CONTROL;
+			cmdq_append(c, item);
+			cmd_list_free(pr->cmdlist);
+			break;
 		}
 
 		free(line);
