@@ -345,12 +345,6 @@ format_draw_centre(struct screen_write_ctx *octx, u_int available, u_int ocx,
 	/* Write left at 0. */
 	format_draw_put(octx, ocx, ocy, left, frs, 0, 0, width_left);
 
-	/* Write after at available - width_after. */
-	format_draw_put(octx, ocx, ocy, after, frs,
-	    available - width_after,
-	    after->cx - width_after,
-	    width_after);
-
 	/* Write right at available - width_right. */
 	format_draw_put(octx, ocx, ocy, right, frs,
 	    available - width_right,
@@ -374,10 +368,10 @@ format_draw_centre(struct screen_write_ctx *octx, u_int available, u_int ocx,
 
 	/*
 	 * Write after at
-	 *     middle + width_list / 2 - width_centre.
+	 *     middle - width_list / 2 + width_list
 	 */
 	format_draw_put(octx, ocx, ocy, after, frs,
-	    middle + width_list / 2,
+	    middle - width_list / 2 + width_list,
 	    0,
 	    width_after);
 
@@ -517,8 +511,9 @@ format_draw(struct screen_write_ctx *octx, const struct grid_cell *base,
 	u_int			 ocx = os->cx, ocy = os->cy, i, width[TOTAL];
 	u_int			 map[] = { LEFT, LEFT, CENTRE, RIGHT };
 	int			 focus_start = -1, focus_end = -1;
-	int			 list_state = -1;
+	int			 list_state = -1, fill = -1;
 	enum style_align	 list_align = STYLE_ALIGN_DEFAULT;
+	struct grid_cell	 gc;
 	struct style		 sy;
 	struct utf8_data	*ud = &sy.gc.data;
 	const char		*cp, *end;
@@ -570,7 +565,7 @@ format_draw(struct screen_write_ctx *octx, const struct grid_cell *base,
 				cp++;
 			}
 
-			/* Draw the cell to th current screen. */
+			/* Draw the cell to the current screen. */
 			screen_write_cell(&ctx[current], &sy.gc);
 			width[current] += ud->width;
 			continue;
@@ -596,6 +591,10 @@ format_draw(struct screen_write_ctx *octx, const struct grid_cell *base,
 		    style_tostring(&sy));
 		free(tmp);
 
+		/* If this style has a fill colour, store it for later. */
+		if (sy.fill != 8)
+			fill = sy.fill;
+
 		/* Check the list state. */
 		switch (sy.list) {
 		case STYLE_LIST_ON:
@@ -717,6 +716,14 @@ format_draw(struct screen_write_ctx *octx, const struct grid_cell *base,
 		    fr->argument, names[fr->index], fr->start, fr->end);
 	}
 
+	/* Clear the available area. */
+	if (fill != -1) {
+		memcpy(&gc, &grid_default_cell, sizeof gc);
+		gc.bg = fill;
+		for (i = 0; i < available; i++)
+			screen_write_putc(octx, &gc, ' ');
+	}
+
 	/*
 	 * Draw the screens. How they are arranged depends on where the list
 	 * appearsq.
@@ -797,7 +804,8 @@ format_width(const char *expanded)
 		} else if (*cp > 0x1f && *cp < 0x7f) {
 			width++;
 			cp++;
-		}
+		} else
+			cp++;
 	}
 	return (width);
 }
