@@ -23,33 +23,10 @@
 
 #include "tmux.h"
 
-struct screen_redraw_ctx {
-	struct client	*c;
-
-	u_int		 lines;
-	int		 top;
-
-	int		 pane_status;
-
-	u_int		 sx;
-	u_int		 sy;
-};
-
-static int	screen_redraw_cell_border1(struct window_pane *, u_int, u_int);
-static int	screen_redraw_cell_border(struct client *, u_int, u_int);
-static int	screen_redraw_check_cell(struct client *, u_int, u_int, int,
-		    struct window_pane **);
-static int	screen_redraw_check_is(u_int, u_int, int, int, struct window *,
-		    struct window_pane *, struct window_pane *);
-
-static int 	screen_redraw_make_pane_status(struct client *, struct window *,
-		    struct window_pane *);
-static void	screen_redraw_draw_pane_status(struct client *, int);
-
 static void	screen_redraw_draw_borders(struct screen_redraw_ctx *);
 static void	screen_redraw_draw_panes(struct screen_redraw_ctx *);
 static void	screen_redraw_draw_status(struct screen_redraw_ctx *);
-static void	screen_redraw_draw_number(struct screen_redraw_ctx *,
+static void	screen_redraw_draw_pane(struct screen_redraw_ctx *,
 		    struct window_pane *);
 
 #define CELL_INSIDE 0
@@ -68,10 +45,6 @@ static void	screen_redraw_draw_number(struct screen_redraw_ctx *,
 
 #define CELL_BORDERS " xqlkmjwvtun~"
 
-#define CELL_STATUS_OFF 0
-#define CELL_STATUS_TOP 1
-#define CELL_STATUS_BOTTOM 2
-
 /* Check if cell is on the border of a particular pane. */
 static int
 screen_redraw_cell_border1(struct window_pane *wp, u_int px, u_int py)
@@ -135,12 +108,12 @@ screen_redraw_check_cell(struct client *c, u_int px, u_int py, int pane_status,
 	if (px > w->sx || py > w->sy)
 		return (CELL_OUTSIDE);
 
-	if (pane_status != CELL_STATUS_OFF) {
+	if (pane_status != PANE_STATUS_OFF) {
 		TAILQ_FOREACH(wp, &w->panes, entry) {
 			if (!window_pane_visible(wp))
 				continue;
 
-			if (pane_status == CELL_STATUS_TOP)
+			if (pane_status == PANE_STATUS_TOP)
 				line = wp->yoff - 1;
 			else
 				line = wp->yoff + wp->sy;
@@ -176,7 +149,7 @@ screen_redraw_check_cell(struct client *c, u_int px, u_int py, int pane_status,
 			borders |= 8;
 		if (px <= w->sx && screen_redraw_cell_border(c, px + 1, py))
 			borders |= 4;
-		if (pane_status == CELL_STATUS_TOP) {
+		if (pane_status == PANE_STATUS_TOP) {
 			if (py != 0 && screen_redraw_cell_border(c, px, py - 1))
 				borders |= 2;
 		} else {
@@ -231,9 +204,9 @@ screen_redraw_check_is(u_int px, u_int py, int type, int pane_status,
 	border = screen_redraw_cell_border1(wantwp, px, py);
 	if (border == 0 || border == -1)
 		return (0);
-	if (pane_status == CELL_STATUS_TOP && border == 4)
+	if (pane_status == PANE_STATUS_TOP && border == 4)
 		return (0);
-	if (pane_status == CELL_STATUS_BOTTOM && border == 3)
+	if (pane_status == PANE_STATUS_BOTTOM && border == 3)
 		return (0);
 
 	/* If there are more than two panes, that's enough. */
@@ -245,7 +218,7 @@ screen_redraw_check_is(u_int px, u_int py, int type, int pane_status,
 		return (1);
 
 	/* With status lines mark the entire line. */
-	if (pane_status != CELL_STATUS_OFF)
+	if (pane_status != PANE_STATUS_OFF)
 		return (1);
 
 	/* Check if the pane covers the whole width. */
@@ -281,8 +254,8 @@ screen_redraw_make_pane_status(struct client *c, struct window *w,
 	struct grid_cell	 gc;
 	const char		*fmt;
 	struct format_tree	*ft;
-	char			*out;
-	size_t			 outlen;
+	char			*expanded;
+	u_int			 width, i;
 	struct screen_write_ctx	 ctx;
 	struct screen		 old;
 
@@ -293,30 +266,33 @@ screen_redraw_make_pane_status(struct client *c, struct window *w,
 
 	fmt = options_get_string(w->options, "pane-border-format");
 
-	ft = format_create(c, NULL, FORMAT_PANE|wp->id, 0);
+	ft = format_create(c, NULL, FORMAT_PANE|wp->id, FORMAT_STATUS);
 	format_defaults(ft, c, NULL, NULL, wp);
 
+	expanded = format_expand_time(ft, fmt);
+	if (wp->sx < 4)
+		wp->status_size = width = 0;
+	else
+		wp->status_size = width = wp->sx - 4;
+
 	memcpy(&old, &wp->status_screen, sizeof old);
-	screen_init(&wp->status_screen, wp->sx, 1, 0);
+	screen_init(&wp->status_screen, width, 1, 0);
 	wp->status_screen.mode = 0;
 
-	out = format_expand(ft, fmt);
-	outlen = screen_write_cstrlen("%s", out);
-	if (outlen > wp->sx - 4)
-		outlen = wp->sx - 4;
-	screen_resize(&wp->status_screen, outlen, 1, 0);
-
 	screen_write_start(&ctx, NULL, &wp->status_screen);
-	screen_write_cursormove(&ctx, 0, 0);
-	screen_write_clearline(&ctx, 8);
-	screen_write_cnputs(&ctx, outlen, &gc, "%s", out);
+
+	gc.attr |= GRID_ATTR_CHARSET;
+	for (i = 0; i < width; i++)
+		screen_write_putc(&ctx, &gc, 'q');
+	gc.attr &= ~GRID_ATTR_CHARSET;
+
+	screen_write_cursormove(&ctx, 0, 0, 0);
+	format_draw(&ctx, &gc, width, expanded, NULL);
 	screen_write_stop(&ctx);
 
-	free(out);
+	free(expanded);
 	format_free(ft);
 
-	wp->status_size = outlen;
-
 	if (grid_compare(wp->status_screen.grid, old.grid) == 0) {
 		screen_free(&old);
 		return (0);
@@ -327,35 +303,67 @@ screen_redraw_make_pane_status(struct client *c, struct window *w,
 
 /* Draw pane status. */
 static void
-screen_redraw_draw_pane_status(struct client *c, int pane_status)
+screen_redraw_draw_pane_status(struct screen_redraw_ctx *ctx)
 {
+	struct client		*c = ctx->c;
 	struct window		*w = c->session->curw->window;
-	struct options		*oo = c->session->options;
 	struct tty		*tty = &c->tty;
 	struct window_pane	*wp;
-	int			 spos;
-	u_int			 yoff;
+	struct screen		*s;
+	u_int			 i, x, width, xoff, yoff, size;
+
+	log_debug("%s: %s @%u", __func__, c->name, w->id);
 
-	spos = options_get_number(oo, "status-position");
 	TAILQ_FOREACH(wp, &w->panes, entry) {
 		if (!window_pane_visible(wp))
 			continue;
-		if (pane_status == CELL_STATUS_TOP)
+		s = &wp->status_screen;
+
+		size = wp->status_size;
+		if (ctx->pane_status == PANE_STATUS_TOP)
 			yoff = wp->yoff - 1;
 		else
 			yoff = wp->yoff + wp->sy;
-		if (spos == 0)
-			yoff += 1;
+		xoff = wp->xoff + 2;
 
-		tty_draw_line(tty, NULL, &wp->status_screen, 0, wp->xoff + 2,
-		    yoff);
+		if (xoff + size <= ctx->ox ||
+		    xoff >= ctx->ox + ctx->sx ||
+		    yoff < ctx->oy ||
+		    yoff >= ctx->oy + ctx->sy)
+			continue;
+
+		if (xoff >= ctx->ox && xoff + size <= ctx->ox + ctx->sx) {
+			/* All visible. */
+			i = 0;
+			x = xoff - ctx->ox;
+			width = size;
+		} else if (xoff < ctx->ox && xoff + size > ctx->ox + ctx->sx) {
+			/* Both left and right not visible. */
+			i = ctx->ox;
+			x = 0;
+			width = ctx->sx;
+		} else if (xoff < ctx->ox) {
+			/* Left not visible. */
+			i = ctx->ox - xoff;
+			x = 0;
+			width = size - i;
+		} else {
+			/* Right not visible. */
+			i = 0;
+			x = xoff - ctx->ox;
+			width = size - x;
+		}
+
+		if (ctx->statustop)
+			yoff += ctx->statuslines;
+		tty_draw_line(tty, NULL, s, i, 0, width, x, yoff - ctx->oy);
 	}
 	tty_cursor(tty, 0, 0);
 }
 
 /* Update status line and change flags if unchanged. */
-void
-screen_redraw_update(struct client *c)
+static int
+screen_redraw_update(struct client *c, int flags)
 {
 	struct window		*w = c->session->curw->window;
 	struct window_pane	*wp;
@@ -368,109 +376,115 @@ screen_redraw_update(struct client *c)
 		redraw = status_prompt_redraw(c);
 	else
 		redraw = status_redraw(c);
-	if (!redraw)
-		c->flags &= ~CLIENT_STATUS;
+	if (!redraw && (~flags & CLIENT_REDRAWSTATUSALWAYS))
+		flags &= ~CLIENT_REDRAWSTATUS;
 
-	if (options_get_number(wo, "pane-border-status") != CELL_STATUS_OFF) {
+	if (c->overlay_draw != NULL)
+		flags |= CLIENT_REDRAWOVERLAY;
+
+	if (options_get_number(wo, "pane-border-status") != PANE_STATUS_OFF) {
 		redraw = 0;
 		TAILQ_FOREACH(wp, &w->panes, entry) {
 			if (screen_redraw_make_pane_status(c, w, wp))
 				redraw = 1;
 		}
 		if (redraw)
-			c->flags |= CLIENT_BORDERS;
+			flags |= CLIENT_REDRAWBORDERS;
 	}
+	return (flags);
+}
+
+/* Set up redraw context. */
+static void
+screen_redraw_set_context(struct client *c, struct screen_redraw_ctx *ctx)
+{
+	struct session	*s = c->session;
+	struct options	*oo = s->options;
+	struct window	*w = s->curw->window;
+	struct options	*wo = w->options;
+	u_int		 lines;
+
+	memset(ctx, 0, sizeof *ctx);
+	ctx->c = c;
+
+	lines = status_line_size(c);
+	if (c->message_string != NULL || c->prompt_string != NULL)
+		lines = (lines == 0) ? 1 : lines;
+	if (lines != 0 && options_get_number(oo, "status-position") == 0)
+		ctx->statustop = 1;
+	ctx->statuslines = lines;
+
+	ctx->pane_status = options_get_number(wo, "pane-border-status");
+
+	tty_window_offset(&c->tty, &ctx->ox, &ctx->oy, &ctx->sx, &ctx->sy);
+
+	log_debug("%s: %s @%u ox=%u oy=%u sx=%u sy=%u %u/%d", __func__, c->name,
+	    w->id, ctx->ox, ctx->oy, ctx->sx, ctx->sy, ctx->statuslines,
+	    ctx->statustop);
 }
 
 /* Redraw entire screen. */
 void
-screen_redraw_screen(struct client *c, int draw_panes, int draw_status,
-    int draw_borders)
+screen_redraw_screen(struct client *c)
 {
-	struct options			*oo = c->session->options;
-	struct tty			*tty = &c->tty;
-	struct window			*w = c->session->curw->window;
-	struct options			*wo = w->options;
-	struct screen_redraw_ctx	 ctx;
+	struct screen_redraw_ctx	ctx;
+	int				flags;
 
 	if (c->flags & CLIENT_SUSPENDED)
 		return;
 
-	memset(&ctx, 0, sizeof ctx);
-	ctx.c = c;
+	flags = screen_redraw_update(c, c->flags);
+	screen_redraw_set_context(c, &ctx);
 
-	if (c->flags & CLIENT_STATUSOFF)
-		ctx.lines = 0;
-	else
-		ctx.lines = status_line_size(c->session);
-	if (c->message_string != NULL || c->prompt_string != NULL)
-		ctx.lines = (ctx.lines == 0) ? 1 : ctx.lines;
-
-	if (ctx.lines != 0 && options_get_number(oo, "status-position") == 0)
-		ctx.top = 1;
-	ctx.pane_status = options_get_number(wo, "pane-border-status");
-
-	ctx.sx = tty->sx;
-	ctx.sy = tty->sy - ctx.lines;
-
-	if (ctx.lines == 0)
-		draw_status = 0;
-
-	if (draw_borders) {
-		if (ctx.pane_status != CELL_STATUS_OFF)
-			screen_redraw_draw_pane_status(c, ctx.pane_status);
+	if (flags & (CLIENT_REDRAWWINDOW|CLIENT_REDRAWBORDERS)) {
+		if (ctx.pane_status != PANE_STATUS_OFF)
+			screen_redraw_draw_pane_status(&ctx);
 		screen_redraw_draw_borders(&ctx);
 	}
-	if (draw_panes)
+	if (flags & CLIENT_REDRAWWINDOW)
 		screen_redraw_draw_panes(&ctx);
-	if (draw_status)
+	if (ctx.statuslines != 0 &&
+	    (flags & (CLIENT_REDRAWSTATUS|CLIENT_REDRAWSTATUSALWAYS)))
 		screen_redraw_draw_status(&ctx);
-	tty_reset(tty);
+	if (c->overlay_draw != NULL && (flags & CLIENT_REDRAWOVERLAY))
+		c->overlay_draw(c, &ctx);
+	tty_reset(&c->tty);
 }
 
-/* Draw a single pane. */
+/* Redraw a single pane. */
 void
 screen_redraw_pane(struct client *c, struct window_pane *wp)
 {
-	u_int	i, yoff;
+	struct screen_redraw_ctx	 ctx;
 
-	if (!window_pane_visible(wp))
+	if (c->overlay_draw != NULL || !window_pane_visible(wp))
 		return;
 
-	yoff = wp->yoff;
-	if (status_at_line(c) == 0)
-		yoff += status_line_size(c->session);
+	screen_redraw_set_context(c, &ctx);
 
-	log_debug("%s: redraw pane %%%u (at %u,%u)", c->name, wp->id,
-	    wp->xoff, yoff);
-
-	for (i = 0; i < wp->sy; i++)
-		tty_draw_pane(&c->tty, wp, i, wp->xoff, yoff);
+	screen_redraw_draw_pane(&ctx, wp);
 	tty_reset(&c->tty);
 }
 
 /* Draw a border cell. */
 static void
-screen_redraw_draw_borders_cell(struct screen_redraw_ctx *ctx, u_int x, u_int y,
-    int small, u_int msgx, u_int msgy, struct grid_cell *m_active_gc,
-    struct grid_cell *active_gc, struct grid_cell *m_other_gc,
-    struct grid_cell *other_gc)
+screen_redraw_draw_borders_cell(struct screen_redraw_ctx *ctx, u_int i, u_int j,
+    struct grid_cell *m_active_gc, struct grid_cell *active_gc,
+    struct grid_cell *m_other_gc, struct grid_cell *other_gc)
 {
-	struct client           *c = ctx->c;
-	struct session          *s = c->session;
-	struct window           *w = s->curw->window;
-	struct tty              *tty = &c->tty;
-	struct window_pane      *wp;
-	struct window_pane      *active = w->active;
-	struct window_pane      *marked = marked_pane.wp;
-	u_int                    type;
-	int                      flag, pane_status = ctx->pane_status;
+	struct client		*c = ctx->c;
+	struct session		*s = c->session;
+	struct window		*w = s->curw->window;
+	struct tty		*tty = &c->tty;
+	struct window_pane	*wp;
+	struct window_pane	*active = w->active;
+	struct window_pane	*marked = marked_pane.wp;
+	u_int			 type, x = ctx->ox + i, y = ctx->oy + j;
+	int			 flag, pane_status = ctx->pane_status;
 
 	type = screen_redraw_check_cell(c, x, y, pane_status, &wp);
 	if (type == CELL_INSIDE)
 		return;
-	if (type == CELL_OUTSIDE && small && x > msgx && y == msgy)
-		return;
 	flag = screen_redraw_check_is(x, y, type, pane_status, w, active, wp);
 
 	if (server_is_marked(s, s->curw, marked_pane.wp) &&
@@ -483,10 +497,10 @@ screen_redraw_draw_borders_cell(struct screen_redraw_ctx *ctx, u_int x, u_int y,
 		tty_attributes(tty, active_gc, NULL);
 	else
 		tty_attributes(tty, other_gc, NULL);
-	if (ctx->top)
-		tty_cursor(tty, x, ctx->lines + y);
+	if (ctx->statustop)
+		tty_cursor(tty, i, ctx->statuslines + j);
 	else
-		tty_cursor(tty, x, y);
+		tty_cursor(tty, i, j);
 	tty_putc(tty, CELL_BORDERS[type]);
 }
 
@@ -497,42 +511,12 @@ screen_redraw_draw_borders(struct screen_redraw_ctx *ctx)
 	struct client		*c = ctx->c;
 	struct session		*s = c->session;
 	struct window		*w = s->curw->window;
-	struct options		*oo = w->options;
 	struct tty		*tty = &c->tty;
+	struct options		*oo = w->options;
 	struct grid_cell	 m_active_gc, active_gc, m_other_gc, other_gc;
-	struct grid_cell	 msg_gc;
-	u_int		 	 i, j, msgx = 0, msgy = 0;
-	int			 small, flags;
-	char			 msg[256];
-	const char		*tmp;
-	size_t			 msglen = 0;
+	u_int		 	 i, j;
 
-	small = (ctx->sy + ctx->top > w->sy) || (ctx->sx > w->sx);
-	if (small) {
-		flags = w->flags & (WINDOW_FORCEWIDTH|WINDOW_FORCEHEIGHT);
-		if (flags == (WINDOW_FORCEWIDTH|WINDOW_FORCEHEIGHT))
-			tmp = "force-width, force-height";
-		else if (flags == WINDOW_FORCEWIDTH)
-			tmp = "force-width";
-		else if (flags == WINDOW_FORCEHEIGHT)
-			tmp = "force-height";
-		else if (c->flags & CLIENT_STATUSOFF)
-			tmp = "status line";
-		else
-			tmp = "a smaller client";
-		xsnprintf(msg, sizeof msg, "(size %ux%u from %s)",
-		    w->sx, w->sy, tmp);
-		msglen = strlen(msg);
-
-		if (ctx->sy - 1 + ctx->top > w->sy && ctx->sx >= msglen) {
-			msgx = ctx->sx - msglen;
-			msgy = ctx->sy - 1 + ctx->top;
-		} else if (ctx->sx - w->sx > msglen) {
-			msgx = ctx->sx - msglen;
-			msgy = ctx->sy - 1 + ctx->top;
-		} else
-			small = 0;
-	}
+	log_debug("%s: %s @%u", __func__, c->name, w->id);
 
 	style_apply(&other_gc, oo, "pane-border-style");
 	style_apply(&active_gc, oo, "pane-active-border-style");
@@ -543,20 +527,12 @@ screen_redraw_draw_borders(struct screen_redraw_ctx *ctx)
 	memcpy(&m_active_gc, &active_gc, sizeof m_active_gc);
 	m_active_gc.attr ^= GRID_ATTR_REVERSE;
 
-	for (j = 0; j < ctx->sy; j++) {
-		for (i = 0; i < ctx->sx; i++) {
-			screen_redraw_draw_borders_cell(ctx, i, j, small,
-			    msgx, msgy, &m_active_gc, &active_gc, &m_other_gc,
-			    &other_gc);
+	for (j = 0; j < tty->sy - ctx->statuslines; j++) {
+		for (i = 0; i < tty->sx; i++) {
+			screen_redraw_draw_borders_cell(ctx, i, j,
+			    &m_active_gc, &active_gc, &m_other_gc, &other_gc);
 		}
 	}
-
-	if (small) {
-		memcpy(&msg_gc, &grid_default_cell, sizeof msg_gc);
-		tty_attributes(tty, &msg_gc, NULL);
-		tty_cursor(tty, msgx, msgy);
-		tty_puts(tty, msg);
-	}
 }
 
 /* Draw the panes. */
@@ -565,21 +541,13 @@ screen_redraw_draw_panes(struct screen_redraw_ctx *ctx)
 {
 	struct client		*c = ctx->c;
 	struct window		*w = c->session->curw->window;
-	struct tty		*tty = &c->tty;
 	struct window_pane	*wp;
-	u_int		 	 i, y;
 
-	if (ctx->top)
-		y = ctx->lines;
-	else
-		y = 0;
+	log_debug("%s: %s @%u", __func__, c->name, w->id);
+
 	TAILQ_FOREACH(wp, &w->panes, entry) {
-		if (!window_pane_visible(wp))
-			continue;
-		for (i = 0; i < wp->sy; i++)
-			tty_draw_pane(tty, wp, i, wp->xoff, y + wp->yoff);
-		if (c->flags & CLIENT_IDENTIFY)
-			screen_redraw_draw_number(ctx, wp);
+		if (window_pane_visible(wp))
+			screen_redraw_draw_pane(ctx, wp);
 	}
 }
 
@@ -588,93 +556,72 @@ static void
 screen_redraw_draw_status(struct screen_redraw_ctx *ctx)
 {
 	struct client	*c = ctx->c;
+	struct window	*w = c->session->curw->window;
 	struct tty	*tty = &c->tty;
+	struct screen	*s = c->status.active;
 	u_int		 i, y;
 
-	if (ctx->top)
+	log_debug("%s: %s @%u", __func__, c->name, w->id);
+
+	if (ctx->statustop)
 		y = 0;
 	else
-		y = ctx->sy;
-	for (i = 0; i < ctx->lines; i++)
-		tty_draw_line(tty, NULL, &c->status.status, i, 0, y);
+		y = c->tty.sy - ctx->statuslines;
+	for (i = 0; i < ctx->statuslines; i++)
+		tty_draw_line(tty, NULL, s, 0, i, UINT_MAX, 0, y + i);
 }
 
-/* Draw number on a pane. */
+/* Draw one pane. */
 static void
-screen_redraw_draw_number(struct screen_redraw_ctx *ctx, struct window_pane *wp)
+screen_redraw_draw_pane(struct screen_redraw_ctx *ctx, struct window_pane *wp)
 {
-	struct client		*c = ctx->c;
-	struct tty		*tty = &c->tty;
-	struct session		*s = c->session;
-	struct options		*oo = s->options;
-	struct window		*w = wp->window;
-	struct grid_cell	 gc;
-	u_int			 idx, px, py, i, j, xoff, yoff;
-	int			 colour, active_colour;
-	char			 buf[16], *ptr;
-	size_t			 len;
+	struct client	*c = ctx->c;
+	struct window	*w = c->session->curw->window;
+	struct tty	*tty = &c->tty;
+	struct screen	*s;
+	u_int		 i, j, top, x, y, width;
 
-	if (window_pane_index(wp, &idx) != 0)
-		fatalx("index not found");
-	len = xsnprintf(buf, sizeof buf, "%u", idx);
+	log_debug("%s: %s @%u %%%u", __func__, c->name, w->id, wp->id);
 
-	if (wp->sx < len)
+	if (wp->xoff + wp->sx <= ctx->ox || wp->xoff >= ctx->ox + ctx->sx)
 		return;
-	colour = options_get_number(oo, "display-panes-colour");
-	active_colour = options_get_number(oo, "display-panes-active-colour");
-
-	px = wp->sx / 2; py = wp->sy / 2;
-	xoff = wp->xoff; yoff = wp->yoff;
-
-	if (ctx->top)
-		yoff += ctx->lines;
-
-	if (wp->sx < len * 6 || wp->sy < 5) {
-		tty_cursor(tty, xoff + px - len / 2, yoff + py);
-		goto draw_text;
-	}
-
-	px -= len * 3;
-	py -= 2;
-
-	memcpy(&gc, &grid_default_cell, sizeof gc);
-	if (w->active == wp)
-		gc.bg = active_colour;
+	if (ctx->statustop)
+		top = ctx->statuslines;
 	else
-		gc.bg = colour;
-	gc.flags |= GRID_FLAG_NOPALETTE;
+		top = 0;
 
-	tty_attributes(tty, &gc, wp);
-	for (ptr = buf; *ptr != '\0'; ptr++) {
-		if (*ptr < '0' || *ptr > '9')
+	s = wp->screen;
+	for (j = 0; j < wp->sy; j++) {
+		if (wp->yoff + j < ctx->oy || wp->yoff + j >= ctx->oy + ctx->sy)
 			continue;
-		idx = *ptr - '0';
+		y = top + wp->yoff + j - ctx->oy;
 
-		for (j = 0; j < 5; j++) {
-			for (i = px; i < px + 5; i++) {
-				tty_cursor(tty, xoff + i, yoff + py + j);
-				if (window_clock_table[idx][j][i - px])
-					tty_putc(tty, ' ');
-			}
+		if (wp->xoff >= ctx->ox &&
+		    wp->xoff + wp->sx <= ctx->ox + ctx->sx) {
+			/* All visible. */
+			i = 0;
+			x = wp->xoff - ctx->ox;
+			width = wp->sx;
+		} else if (wp->xoff < ctx->ox &&
+		    wp->xoff + wp->sx > ctx->ox + ctx->sx) {
+			/* Both left and right not visible. */
+			i = ctx->ox;
+			x = 0;
+			width = ctx->sx;
+		} else if (wp->xoff < ctx->ox) {
+			/* Left not visible. */
+			i = ctx->ox - wp->xoff;
+			x = 0;
+			width = wp->sx - i;
+		} else {
+			/* Right not visible. */
+			i = 0;
+			x = wp->xoff - ctx->ox;
+			width = ctx->sx - x;
 		}
-		px += 6;
+		log_debug("%s: %s %%%u line %u,%u at %u,%u, width %u",
+		    __func__, c->name, wp->id, i, j, x, y, width);
+
+		tty_draw_line(tty, wp, s, i, j, width, x, y);
 	}
-
-	len = xsnprintf(buf, sizeof buf, "%ux%u", wp->sx, wp->sy);
-	if (wp->sx < len || wp->sy < 6)
-		return;
-	tty_cursor(tty, xoff + wp->sx - len, yoff);
-
-draw_text:
-	memcpy(&gc, &grid_default_cell, sizeof gc);
-	if (w->active == wp)
-		gc.fg = active_colour;
-	else
-		gc.fg = colour;
-	gc.flags |= GRID_FLAG_NOPALETTE;
-
-	tty_attributes(tty, &gc, wp);
-	tty_puts(tty, buf);
-
-	tty_cursor(tty, 0, 0);
 }
