@@ -30,8 +30,9 @@
 
 /* Default style. */
 static struct style style_default = {
-	{ 0, 0, 8, 8, { { ' ' }, 0, 1, 1 } },
+	{ { { ' ' }, 0, 1, 1 }, 0, 0, 8, 8, 0  },
 
+	8,
 	STYLE_ALIGN_DEFAULT,
 	STYLE_LIST_OFF,
 
@@ -57,10 +58,8 @@ style_parse(struct style *sy, const struct grid_cell *base, const char *in)
 	style_copy(&saved, sy);
 
 	do {
-		while (*in != '\0' && strchr(delimiters, *in) != NULL) {
+		while (*in != '\0' && strchr(delimiters, *in) != NULL)
 			in++;
-			end--;
-		}
 		if (*in == '\0')
 			break;
 
@@ -129,6 +128,10 @@ style_parse(struct style *sy, const struct grid_cell *base, const char *in)
 				sy->align = STYLE_ALIGN_RIGHT;
 			else
 				goto error;
+		} else if (end > 5 && strncasecmp(tmp, "fill=", 5) == 0) {
+			if ((value = colour_fromstring(tmp + 5)) == -1)
+				goto error;
+			sy->fill = value;
 		} else if (end > 3 && strncasecmp(tmp + 1, "g=", 2) == 0) {
 			if ((value = colour_fromstring(tmp + 3)) == -1)
 				goto error;
@@ -172,7 +175,7 @@ style_tostring(struct style *sy)
 {
 	struct grid_cell	*gc = &sy->gc;
 	int			 off = 0;
-	const char		*comma = "", *tmp;
+	const char		*comma = "", *tmp = "";
 	static char		 s[256];
 	char			 b[16];
 
@@ -215,6 +218,11 @@ style_tostring(struct style *sy)
 		    tmp);
 		comma = ",";
 	}
+	if (sy->fill != 8) {
+		off += xsnprintf(s + off, sizeof s - off, "%sfill=%s", comma,
+		    colour_tostring(sy->fill));
+		comma = ",";
+	}
 	if (gc->fg != 8) {
 		off += xsnprintf(s + off, sizeof s - off, "%sfg=%s", comma,
 		    colour_tostring(gc->fg));
@@ -292,6 +300,8 @@ style_equal(struct style *sy1, struct style *sy2)
 		return (0);
 	if ((gc1->attr & STYLE_ATTR_MASK) != (gc2->attr & STYLE_ATTR_MASK))
 		return (0);
+	if (sy1->fill != sy2->fill)
+		return (0);
 	if (sy1->align != sy2->align)
 		return (0);
 	return (1);
