@@ -18,6 +18,7 @@
 
 #include <sys/types.h>
 
+#include <ctype.h>
 #include <errno.h>
 #include <stdlib.h>
 #include <string.h>
@@ -207,7 +208,13 @@ utf8_strvis(char *dst, const char *src, size_t len, int flag)
 			/* Not a complete, valid UTF-8 character. */
 			src -= ud.have;
 		}
-		if (src < end - 1)
+		if (src[0] == '$' && src < end - 1) {
+			if (isalpha((u_char)src[1]) ||
+			    src[1] == '_' ||
+			    src[1] == '{')
+				*dst++ = '\\';
+			*dst++ = '$';
+		} else if (src < end - 1)
 			dst = vis(dst, src[0], flag, src[1]);
 		else if (src < end)
 			dst = vis(dst, src[0], flag, '\0');
@@ -232,6 +239,30 @@ utf8_stravis(char **dst, const char *src, int flag)
 	return (len);
 }
 
+/* Does this string contain anything that isn't valid UTF-8? */
+int
+utf8_isvalid(const char *s)
+{
+	struct utf8_data	 ud;
+	const char		*end;
+	enum utf8_state		 more;
+
+	end = s + strlen(s);
+	while (s < end) {
+		if ((more = utf8_open(&ud, *s)) == UTF8_MORE) {
+			while (++s < end && more == UTF8_MORE)
+				more = utf8_append(&ud, *s);
+			if (more == UTF8_DONE)
+				continue;
+			return (0);
+		}
+		if (*s < 0x20 || *s > 0x7e)
+			return (0);
+		s++;
+	}
+	return (1);
+}
+
 /*
  * Sanitize a string, changing any UTF-8 characters to '_'. Caller should free
  * the returned string. Anything not valid printable ASCII or UTF-8 is
@@ -384,67 +415,6 @@ utf8_cstrwidth(const char *s)
 	return (width);
 }
 
-/* Trim UTF-8 string to width. Caller frees. */
-char *
-utf8_trimcstr(const char *s, u_int width)
-{
-	struct utf8_data	*tmp, *next;
-	char			*out;
-	u_int			 at;
-
-	tmp = utf8_fromcstr(s);
-
-	at = 0;
-	for (next = tmp; next->size != 0; next++) {
-		if (at + next->width > width) {
-			next->size = 0;
-			break;
-		}
-		at += next->width;
-	}
-
-	out = utf8_tocstr(tmp);
-	free(tmp);
-	return (out);
-}
-
-/* Trim UTF-8 string to width. Caller frees. */
-char *
-utf8_rtrimcstr(const char *s, u_int width)
-{
-	struct utf8_data	*tmp, *next, *end;
-	char			*out;
-	u_int			 at;
-
-	tmp = utf8_fromcstr(s);
-
-	for (end = tmp; end->size != 0; end++)
-		/* nothing */;
-	if (end == tmp) {
-		free(tmp);
-		return (xstrdup(""));
-	}
-	next = end - 1;
-
-	at = 0;
-	for (;;)
-	{
-		if (at + next->width > width) {
-			next++;
-			break;
-		}
-		at += next->width;
-
-		if (next == tmp)
-			break;
-		next--;
-	}
-
-	out = utf8_tocstr(next);
-	free(tmp);
-	return (out);
-}
-
 /* Pad UTF-8 string to width. Caller frees. */
 char *
 utf8_padcstr(const char *s, u_int width)
@@ -465,3 +435,23 @@ utf8_padcstr(const char *s, u_int width)
 	out[slen] = '\0';
 	return (out);
 }
+
+int
+utf8_cstrhas(const char *s, const struct utf8_data *ud)
+{
+	struct utf8_data	*copy, *loop;
+	int			 found = 0;
+
+	copy = utf8_fromcstr(s);
+	for (loop = copy; loop->size != 0; loop++) {
+		if (loop->size != ud->size)
+			continue;
+		if (memcmp(loop->data, ud->data, loop->size) == 0) {
+			found = 1;
+			break;
+		}
+	}
+	free(copy);
+
+	return (found);
+}
