@@ -38,6 +38,22 @@ static void		 window_buffer_key(struct window_mode_entry *,
 #define WINDOW_BUFFER_DEFAULT_FORMAT \
 	"#{buffer_size} bytes (#{t:buffer_created})"
 
+static const struct menu_item window_buffer_menu_items[] = {
+	{ "Paste", 'p', NULL },
+	{ "Paste Tagged", 'P', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Tag", 't', NULL },
+	{ "Tag All", '\024', NULL },
+	{ "Tag None", 'T', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Delete", 'd', NULL },
+	{ "Delete Tagged", 'D', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Cancel", 'q', NULL },
+
+	{ NULL, KEYC_NONE, NULL }
+};
+
 const struct window_mode window_buffer_mode = {
 	.name = "buffer-mode",
 	.default_format = WINDOW_BUFFER_DEFAULT_FORMAT,
@@ -66,6 +82,9 @@ struct window_buffer_itemdata {
 };
 
 struct window_buffer_modedata {
+	struct window_pane		 *wp;
+	struct cmd_find_state		  fs;
+
 	struct mode_tree_data		 *data;
 	char				 *command;
 	char				 *format;
@@ -137,6 +156,9 @@ window_buffer_build(void *modedata, u_int sort_type, __unused uint64_t *tag,
 	struct paste_buffer		*pb;
 	char				*text, *cp;
 	struct format_tree		*ft;
+	struct session			*s = NULL;
+	struct winlink			*wl = NULL;
+	struct window_pane		*wp = NULL;
 
 	for (i = 0; i < data->item_size; i++)
 		window_buffer_free_item(data->item_list[i]);
@@ -167,6 +189,12 @@ window_buffer_build(void *modedata, u_int sort_type, __unused uint64_t *tag,
 		break;
 	}
 
+	if (cmd_find_valid_state(&data->fs)) {
+		s = data->fs.s;
+		wl = data->fs.wl;
+		wp = data->fs.wp;
+	}
+
 	for (i = 0; i < data->item_size; i++) {
 		item = data->item_list[i];
 
@@ -174,6 +202,7 @@ window_buffer_build(void *modedata, u_int sort_type, __unused uint64_t *tag,
 		if (pb == NULL)
 			continue;
 		ft = format_create(NULL, NULL, FORMAT_NONE, 0);
+		format_defaults(ft, NULL, s, wl, wp);
 		format_defaults_paste_buffer(ft, pb);
 
 		if (filter != NULL) {
@@ -216,7 +245,7 @@ window_buffer_draw(__unused void *modedata, void *itemdata,
 		at = 0;
 		while (end != pdata + psize && *end != '\n') {
 			if ((sizeof line) - at > 5) {
-				cp = vis(line + at, *end, VIS_TAB|VIS_OCTAL, 0);
+				cp = vis(line + at, *end, VIS_OCTAL|VIS_TAB, 0);
 				at = cp - line;
 			}
 			end++;
@@ -252,15 +281,30 @@ window_buffer_search(__unused void *modedata, void *itemdata, const char *ss)
 	return (memmem(bufdata, bufsize, ss, strlen(ss)) != NULL);
 }
 
+static void
+window_buffer_menu(void *modedata, struct client *c, key_code key)
+{
+	struct window_buffer_modedata	*data = modedata;
+	struct window_pane		*wp = data->wp;
+	struct window_mode_entry	*wme;
+
+	wme = TAILQ_FIRST(&wp->modes);
+	if (wme == NULL || wme->data != modedata)
+		return;
+	window_buffer_key(wme, c, NULL, NULL, key, NULL);
+}
+
 static struct screen *
-window_buffer_init(struct window_mode_entry *wme,
-    __unused struct cmd_find_state *fs, struct args *args)
+window_buffer_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
+    struct args *args)
 {
 	struct window_pane		*wp = wme->wp;
 	struct window_buffer_modedata	*data;
 	struct screen			*s;
 
 	wme->data = data = xcalloc(1, sizeof *data);
+	data->wp = wp;
+	cmd_find_copy_state(&data->fs, fs);
 
 	if (args == NULL || !args_has(args, 'F'))
 		data->format = xstrdup(WINDOW_BUFFER_DEFAULT_FORMAT);
@@ -272,8 +316,9 @@ window_buffer_init(struct window_mode_entry *wme,
 		data->command = xstrdup(args->argv[0]);
 
 	data->data = mode_tree_start(wp, args, window_buffer_build,
-	    window_buffer_draw, window_buffer_search, data,
-	    window_buffer_sort_list, nitems(window_buffer_sort_list), &s);
+	    window_buffer_draw, window_buffer_search, window_buffer_menu, data,
+	    window_buffer_menu_items, window_buffer_sort_list,
+	    nitems(window_buffer_sort_list), &s);
 	mode_tree_zoom(data->data, args);
 
 	mode_tree_build(data->data);
