@@ -53,6 +53,22 @@ static void		 window_tree_key(struct window_mode_entry *,
 		"}" \
 	"}"
 
+static const struct menu_item window_tree_menu_items[] = {
+	{ "Select", 'E', NULL },
+	{ "Expand", 'R', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Tag", 't', NULL },
+	{ "Tag All", '\024', NULL },
+	{ "Tag None", 'T', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Kill", 'x', NULL },
+	{ "Kill Tagged", 'X', NULL },
+	{ "", KEYC_NONE, NULL },
+	{ "Cancel", 'q', NULL },
+
+	{ NULL, KEYC_NONE, NULL }
+};
+
 const struct window_mode window_tree_mode = {
 	.name = "tree-mode",
 	.default_format = WINDOW_TREE_DEFAULT_FORMAT,
@@ -785,7 +801,8 @@ window_tree_search(__unused void *modedata, void *itemdata, const char *ss)
 	struct session			*s;
 	struct winlink			*wl;
 	struct window_pane		*wp;
-	const char			*cmd;
+	char				*cmd;
+	int				 retval;
 
 	window_tree_pull_item(item, &s, &wl, &wp);
 
@@ -806,11 +823,26 @@ window_tree_search(__unused void *modedata, void *itemdata, const char *ss)
 		cmd = osdep_get_name(wp->fd, wp->tty);
 		if (cmd == NULL || *cmd == '\0')
 			return (0);
-		return (strstr(cmd, ss) != NULL);
+		retval = (strstr(cmd, ss) != NULL);
+		free(cmd);
+		return retval;
 	}
 	return (0);
 }
 
+static void
+window_tree_menu(void *modedata, struct client *c, key_code key)
+{
+	struct window_tree_modedata	*data = modedata;
+	struct window_pane		*wp = data->wp;
+	struct window_mode_entry	*wme;
+
+	wme = TAILQ_FIRST(&wp->modes);
+	if (wme == NULL || wme->data != modedata)
+		return;
+	window_tree_key(wme, c, NULL, NULL, key, NULL);
+}
+
 static struct screen *
 window_tree_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
     struct args *args)
@@ -820,6 +852,8 @@ window_tree_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
 	struct screen			*s;
 
 	wme->data = data = xcalloc(1, sizeof *data);
+	data->wp = wp;
+	data->references = 1;
 
 	if (args_has(args, 's'))
 		data->type = WINDOW_TREE_SESSION;
@@ -829,9 +863,6 @@ window_tree_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
 		data->type = WINDOW_TREE_PANE;
 	memcpy(&data->fs, fs, sizeof data->fs);
 
-	data->wp = wp;
-	data->references = 1;
-
 	if (args == NULL || !args_has(args, 'F'))
 		data->format = xstrdup(WINDOW_TREE_DEFAULT_FORMAT);
 	else
@@ -843,7 +874,8 @@ window_tree_init(struct window_mode_entry *wme, struct cmd_find_state *fs,
 	data->squash_groups = !args_has(args, 'G');
 
 	data->data = mode_tree_start(wp, args, window_tree_build,
-	    window_tree_draw, window_tree_search, data, window_tree_sort_list,
+	    window_tree_draw, window_tree_search, window_tree_menu, data,
+	    window_tree_menu_items, window_tree_sort_list,
 	    nitems(window_tree_sort_list), &s);
 	mode_tree_zoom(data->data, args);
 
@@ -1006,7 +1038,7 @@ window_tree_kill_each(__unused void* modedata, void* itemdata,
 	case WINDOW_TREE_SESSION:
 		if (s != NULL) {
 			server_destroy_session(s);
-			session_destroy(s, __func__);
+			session_destroy(s, 1, __func__);
 		}
 		break;
 	case WINDOW_TREE_WINDOW:
@@ -1143,7 +1175,7 @@ window_tree_key(struct window_mode_entry *wme, struct client *c,
 		item = new_item;
 		data->offset = 0;
 	}
-	if (KEYC_IS_MOUSE(key))
+	if (KEYC_IS_MOUSE(key) && m != NULL)
 		key = window_tree_mouse(data, key, x, item);
 	switch (key) {
 	case '<':
@@ -1204,7 +1236,6 @@ window_tree_key(struct window_mode_entry *wme, struct client *c,
 		free(prompt);
 		break;
 	case '\r':
-		item = mode_tree_get_current(data->data);
 		name = window_tree_get_target(item, &fs);
 		if (name != NULL)
 			mode_tree_run_command(c, NULL, data->command, name);
